#include <kos.h>
#include <dc/g2bus.h>
#include "arm/aica_cmd_iface.h"
#include "dc_sound.h"

#define SPU_RAM_BASE            0xa0800000

uint32 aica_ram_top;

struct {
  uint32 address;
  uint32 size;
} channelBuffers[32];

extern uint8 snd_stream_drv[];
extern uint8 snd_stream_drv_end[];

/* "Kicks" the aica to process commands */
static void kick_aica(void) {
        g2_write_32(SPU_RAM_BASE + AICA_CMD, AICA_CMD_KICK);
        g2_fifo_wait();
        /* wait for the aica */
        while (1) {
                int res;

                res=g2_read_32(SPU_RAM_BASE + AICA_CMD);
                g2_fifo_wait();
                if (!res)
                        return;
        }
}

void dc_sound_playChannel(int channel) {

  aica_channel ch;

  spu_memread(&ch, AICA_CHANNEL(channel), sizeof(ch));
  ch.cmd = AICA_CMD_START;
  spu_memload(AICA_CHANNEL(channel), &ch, sizeof(ch));
  kick_aica();
}

void dc_sound_stopChannel(int channel) {
  aica_channel ch;

  spu_memread(&ch, AICA_CHANNEL(channel), sizeof(ch));
  ch.cmd = AICA_CMD_STOP;
  spu_memload(AICA_CHANNEL(channel), &ch, sizeof(ch));
  kick_aica();
}

void dc_sound_clearChannel(int channel) {
  spu_memset(channelBuffers[channel].address, 0, channelBuffers[channel].size);
}

void dc_sound_initChannel(int channel, aica_channel *ch) {

  ch->cmd = AICA_CMD_START;
  ch->pos = channelBuffers[channel].address;
  spu_memload(AICA_CHANNEL(channel), ch, sizeof(aica_channel));
  kick_aica();
}

void dc_sound_writeData( int channel, int offset, void *pdata, int size){
  spu_memload( channelBuffers[channel].address, pdata, size );
}

void dc_sound_setChannelFreq(int channel, int freq) {

  aica_channel ch;

  spu_memread(&ch, AICA_CHANNEL(channel), sizeof(ch));
  ch.cmd = AICA_CMD_UPDATE | AICA_UPDATE_SET_FREQ;
  ch.freq = freq;
  spu_memload(AICA_CHANNEL(channel), &ch, sizeof(ch));
  kick_aica();
}

void dc_sound_setChannelVolume(int channel, int vol){
  aica_channel ch;

  spu_memread(&ch, AICA_CHANNEL(channel), sizeof(ch));
  ch.cmd = AICA_CMD_UPDATE | AICA_UPDATE_SET_VOL;
  ch.vol = vol;
  spu_memload(AICA_CHANNEL(channel), &ch, sizeof(ch));
  kick_aica();
}

void dc_sound_setChannelPanning(int channel, int pan){
  aica_channel ch;

  spu_memread(&ch, AICA_CHANNEL(channel), sizeof(ch));
  ch.cmd = AICA_CMD_UPDATE | AICA_UPDATE_SET_PAN;
  ch.pan = pan;
  spu_memload(AICA_CHANNEL(channel), &ch, sizeof(ch));
  kick_aica();

}

uint32 aica_malloc(int size) {
  uint32 ret = aica_ram_top;
  aica_ram_top += size;
  return ret;
}

/* Create a sample buffer for a channel in AICA ram */

void dc_sound_createChannel( int channel, int size ) {
  channelBuffers[channel].address = aica_malloc(size);
  channelBuffers[channel].size = size;
}

void dc_sound_destroyChannels() {
  aica_ram_top = AICA_RAM_START;
}

void dc_sound_init() {
  int drvsize;

  spu_disable();
  spu_memset(0,0,0x031000);
  drvsize = snd_stream_drv_end - snd_stream_drv;
  if (drvsize & ~3)
    drvsize = (drvsize + 4) & ~3;
  spu_memload(0, snd_stream_drv, drvsize); 
  spu_enable();

  timer_spin_sleep(10);

  dc_sound_destroyChannels();
}
